/*
 * Copyright 2004-2007 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* System Timer Interrupt reconfigured to run in free-run mode.
 * Author: Vitaly Wool
 * Copyright 2004 MontaVista Software Inc.
 */

/*!
 * @file time.c
 * @brief This file contains OS tick implementations.
 *
 * This file contains OS tick implementations.
 *
 * @ingroup Timers
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/io.h>
#include <asm/hardware.h>
#include <asm/mach/time.h>
#include <asm/arch/common.h>

/* OS tick defines */
#define MXC_GPT_INT_TICK		MXC_INT_GPT
#define MXC_GPT_TCMP_TICK		MXC_GPT_TCMP(MXC_TIMER_GPT1)
#define MXC_GPT_TSTAT_TICK		MXC_GPT_TSTAT(MXC_TIMER_GPT1)
#define MXC_GPT_TCTL_TICK		MXC_GPT_TCTL(MXC_TIMER_GPT1)
#define MXC_GPT_TPRER_TICK		MXC_GPT_TPRER(MXC_TIMER_GPT1)
#define MXC_GPT_TCN_TICK		MXC_GPT_TCN(MXC_TIMER_GPT1)

extern unsigned long clk_early_get_timer_rate(void);

/*!
 * This is the timer interrupt service routine to do required tasks.
 *
 * @param  irq          GPT interrupt source number (not used)
 * @param  dev_id       this parameter is not used
 * @return always returns \b IRQ_HANDLED as defined in
 *         include/linux/interrupt.h.
 */
static irqreturn_t mxc_timer_interrupt(int irq, void *dev_id,
		struct pt_regs *regs)
{
	unsigned int next_match;

	write_seqlock(&xtime_lock);

	do {
		timer_tick(regs);
		next_match = __raw_readl(MXC_GPT_TCMP_TICK) + LATCH;
		__raw_writel(GPT_TSTAT_COMP, MXC_GPT_TSTAT_TICK);
		__raw_writel(next_match, MXC_GPT_TCMP_TICK);
	} while ((signed long)(next_match - __raw_readl(MXC_GPT_TCN_TICK)) <=
		 0);

	write_sequnlock(&xtime_lock);

	return IRQ_HANDLED;
}

/*!
 * The OS tick timer interrupt structure.
 */
static struct irqaction timer_irq = {
	.name    = "MXC Timer Tick",
	.flags   = IRQF_DISABLED | IRQF_TIMER,
	.handler = mxc_timer_interrupt,
};

/*!
 * This function is used to initialize the GPT to produce an interrupt
 * every 10 msec. It is called by the start_kernel() during system startup.
 */
void mxc_init_time(void)
{
	u32 reg, v;

	__raw_writel(0, MXC_GPT_TCTL_TICK);
	__raw_writel(GPT_TCTL_SWR, MXC_GPT_TCTL_TICK);

	while ((__raw_readl(MXC_GPT_TCTL_TICK) & GPT_TCTL_SWR) != 0)
		cpu_relax();

	reg = GPT_TCTL_FRR | GPT_TCTL_COMPEN | GPT_TCTL_SRC_PER1;

	__raw_writel(reg, MXC_GPT_TCTL_TICK);

	v = clk_early_get_timer_rate();

	if ((v % CLOCK_TICK_RATE) != 0) {
		pr_info("\nWARNING: Can't generate CLOCK_TICK_RATE at %d Hz\n",
			CLOCK_TICK_RATE);
	}
	pr_info("Actual CLOCK_TICK_RATE is %d Hz\n",
		v / ((__raw_readl(MXC_GPT_TPRER_TICK) & 0x7FF) + 1));

	reg = __raw_readl(MXC_GPT_TCN_TICK);
	reg += LATCH;
	__raw_writel(reg, MXC_GPT_TCMP_TICK);

	setup_irq(MXC_GPT_INT_TICK, &timer_irq);

	reg = __raw_readl(MXC_GPT_TCTL_TICK) | GPT_TCTL_TEN;
	__raw_writel(reg, MXC_GPT_TCTL_TICK);
}

struct sys_timer mxc_timer = {
	.init = mxc_init_time,
};
